import datetime as dt
from dateutil.relativedelta import relativedelta
import numpy as np

import pandas as pds
# Orbits period is a pandas.Timedelta kwarg, and the pandas repr
# does not include a module name. Import required to run eval
# on Orbit representation
from pandas import Timedelta  # noqa: F401
import pytest

import pysat


class TestOrbitsUserInterface():
    def setup(self):
        """ Set up User Interface unit tests
        """
        self.in_args = ['pysat', 'testing']
        self.in_kwargs = {'clean_level': 'clean', 'update_files': True}
        self.testInst = None
        self.stime = dt.datetime(2009, 1, 1)

    def teardown(self):
        """ Tear down user interface tests
        """
        del self.in_args, self.in_kwargs, self.testInst, self.stime

    def test_orbit_w_bad_kind(self):
        """ Test orbit failure with bad 'kind' input
        """
        self.in_kwargs['orbit_info'] = {'index': 'mlt', 'kind': 'cats'}
        with pytest.raises(ValueError):
            self.testInst = pysat.Instrument(*self.in_args, **self.in_kwargs)

    @pytest.mark.parametrize("info", [({'index': 'magnetic local time',
                                        'kind': 'longitude'}),
                                      (None),
                                      ({'index': 'magnetic local time',
                                        'kind': 'lt'}),
                                      ({'index': 'magnetic local time',
                                       'kind': 'polar'}),
                                      ({'index': 'magnetic local time',
                                        'kind': 'orbit'})])
    def test_orbit_w_bad_orbit_info(self, info):
        """ Test orbit failure on iteration with orbit initialization
        """
        self.in_kwargs['orbit_info'] = info
        self.testInst = pysat.Instrument(*self.in_args, **self.in_kwargs)
        self.testInst.load(date=self.stime)

        with pytest.raises(ValueError):
            self.testInst.orbits.next()

    @pytest.mark.parametrize("info", [({'index': 'magnetic local time',
                                        'kind': 'polar'}),
                                      ({'index': 'magnetic local time',
                                        'kind': 'orbit'}),
                                      ({'index': 'magnetic local time',
                                        'kind': 'longitude'}),
                                      ({'index': 'magnetic local time',
                                        'kind': 'lt'})])
    def test_orbit_polar_w_missing_orbit_index(self, info):
        """ Test orbit failure oon iteration with missing orbit index
        """
        self.in_kwargs['orbit_info'] = info
        self.testInst = pysat.Instrument(*self.in_args, **self.in_kwargs)

        # Force index to None beforee loading and iterating
        self.testInst.orbits.orbit_index = None
        self.testInst.load(date=self.stime)
        with pytest.raises(ValueError):
            self.testInst.orbits.next()

    def test_orbit_repr(self):
        """ Test the Orbit representation
        """
        self.in_kwargs['orbit_info'] = {'index': 'mlt'}
        self.testInst = pysat.Instrument(*self.in_args, **self.in_kwargs)
        out_str = self.testInst.orbits.__repr__()

        assert out_str.find("Orbits(") >= 0

    def test_orbit_str(self):
        """ Test the Orbit string representation with data
        """
        self.in_kwargs['orbit_info'] = {'index': 'mlt'}
        self.testInst = pysat.Instrument(*self.in_args, **self.in_kwargs)
        self.testInst.load(date=self.stime)
        out_str = self.testInst.orbits.__str__()

        assert out_str.find("Orbit Settings") >= 0
        assert out_str.find("Orbit Lind: local time") < 0


class TestSpecificUTOrbits():
    def setup(self):
        """Runs before every method to create a clean testing setup
        """
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']
        self.inc_min = 97
        self.etime = None

    def teardown(self):
        """Runs after every method to clean up previous testing
        """
        del self.testInst, self.stime, self.inc_min, self.etime

    @pytest.mark.parametrize('orbit_inc', [(0), (1), (-1), (-2), (14)])
    def test_single_orbit_call_by_index(self, orbit_inc):
        """Test successful orbit call by index
        """
        # Load the data
        self.testInst.load(date=self.stime)
        self.testInst.orbits[orbit_inc]

        # Increment the time
        if orbit_inc >= 0:
            self.stime += dt.timedelta(minutes=orbit_inc * self.inc_min)
        else:
            self.stime += dt.timedelta(minutes=self.inc_min
                                       * (np.ceil(1440.0 / self.inc_min)
                                          + orbit_inc))
        self.etime = self.stime + dt.timedelta(seconds=(self.inc_min * 60 - 1))

        # Test the time
        assert (self.testInst.index[0] == self.stime)
        assert (self.testInst.index[-1] == self.etime)

    @pytest.mark.parametrize("orbit_ind,raise_err", [(17, Exception),
                                                     (None, TypeError)])
    def test_single_orbit_call_bad_index(self, orbit_ind, raise_err):
        """ Test orbit failure with bad index
        """
        self.testInst.load(date=self.stime)
        with pytest.raises(raise_err):
            self.testInst.orbits[orbit_ind]

    def test_oribt_number_via_current_multiple_orbit_calls_in_day(self):
        """ Test orbit number with mulitple orbits calls in a day
        """
        self.testInst.load(date=self.stime)
        self.testInst.bounds = (self.stime, None)
        true_vals = np.arange(15)
        true_vals[-1] = 0
        test_vals = []
        for i, inst in enumerate(self.testInst.orbits):
            if i > 14:
                break
            test_vals.append(self.testInst.orbits.current)

        assert np.all(test_vals == true_vals)

    def test_all_single_orbit_calls_in_day(self):
        """ Test all single orbit calls in a day
        """
        self.testInst.load(date=self.stime)
        self.testInst.bounds = (self.stime, None)
        for i, inst in enumerate(self.testInst.orbits):
            if i > 14:
                break

            # Test the start index
            self.etime = self.stime + i * relativedelta(minutes=self.inc_min)
            assert self.testInst.index[0] == self.etime

            # Test the end index
            self.etime += relativedelta(seconds=((self.inc_min * 60) - 1))
            assert self.testInst.index[-1] == self.etime

    def test_orbit_next_call_no_loaded_data(self):
        """ Test orbit next call without loading data
        """
        self.testInst.orbits.next()
        assert (self.testInst.index[0] == dt.datetime(2008, 1, 1))
        assert (self.testInst.index[-1] == dt.datetime(2008, 1, 1, 0, 38, 59))

    def test_orbit_prev_call_no_loaded_data(self):
        """ Test orbit previous call without loading data
        """
        self.testInst.orbits.prev()
        # this isn't a full orbit
        assert (self.testInst.index[-1]
                == dt.datetime(2010, 12, 31, 23, 59, 59))
        assert (self.testInst.index[0] == dt.datetime(2010, 12, 31, 23, 49))

    def test_single_orbit_call_orbit_starts_0_UT_using_next(self):
        """ Test orbit next call with data
        """
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        self.etime = self.stime + dt.timedelta(seconds=(self.inc_min * 60 - 1))
        assert (self.testInst.index[0] == self.stime)
        assert (self.testInst.index[-1] == self.etime)

    def test_single_orbit_call_orbit_starts_0_UT_using_prev(self):
        """ Test orbit prev call with data
        """
        self.testInst.load(date=self.stime)
        self.testInst.orbits.prev()
        self.stime += 14 * relativedelta(minutes=self.inc_min)
        self.etime = self.stime + dt.timedelta(seconds=((self.inc_min * 60)
                                                        - 1))
        assert self.testInst.index[0] == self.stime
        assert self.testInst.index[-1] == self.etime

    def test_single_orbit_call_orbit_starts_off_0_UT_using_next(self):
        """ Test orbit next call with data for previous day
        """
        self.stime -= dt.timedelta(days=1)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        assert (self.testInst.index[0] == dt.datetime(2008, 12, 30, 23, 45))
        assert (self.testInst.index[-1]
                == (dt.datetime(2008, 12, 30, 23, 45)
                    + relativedelta(seconds=(self.inc_min * 60 - 1))))

    def test_single_orbit_call_orbit_starts_off_0_UT_using_prev(self):
        self.stime -= dt.timedelta(days=1)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.prev()
        assert (self.testInst.index[0]
                == (dt.datetime(2009, 1, 1)
                    - relativedelta(minutes=self.inc_min)))
        assert (self.testInst.index[-1]
                == (dt.datetime(2009, 1, 1) - relativedelta(seconds=1)))


class TestGeneralOrbitsMLT():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']
        return

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime
        return

    def test_equality_with_copy(self):
        """Test that copy is the same as original"""
        self.out = self.testInst.orbits.copy()
        assert self.out == self.testInst.orbits
        return

    def test_equality_with_data_with_copy(self):
        """Test that copy is the same as original"""
        # Load data
        self.testInst.load(date=self.stime)

        # Load up an orbit
        self.testInst.orbits[0]
        self.out = self.testInst.orbits.copy()

        assert self.out == self.testInst.orbits
        return

    def test_inequality_different_data(self):
        """Test that equality is false if different data"""
        # Load data
        self.testInst.load(date=self.stime)

        # Load up an orbit
        self.testInst.orbits[0]

        # Make copy
        self.out = self.testInst.orbits.copy()

        # Modify data
        self.out._full_day_data = self.testInst._null_data

        assert self.out != self.testInst.orbits
        return

    def test_inequality_modified_object(self):
        """Test that equality is false if other missing attributes"""
        self.out = self.testInst.orbits.copy()

        # Remove attribute
        del self.out.orbit_index

        assert self.testInst.orbits != self.out
        return

    def test_inequality_reduced_object(self):
        """Test that equality is false if self missing attributes"""
        self.out = self.testInst.orbits.copy()
        self.out.hi_there = 'hi'
        assert self.testInst.orbits != self.out
        return

    def test_inequality_different_type(self):
        """Test that equality is false if different type"""
        assert self.testInst.orbits != self.testInst
        return

    def test_eval_repr(self):
        """Test eval of repr recreates object"""
        # eval and repr don't play nice for custom functions
        if len(self.testInst.custom_functions) != 0:
            self.testInst.custom_clear()

        self.out = eval(self.testInst.orbits.__repr__())
        assert self.out == self.testInst.orbits
        return

    def test_repr_and_copy(self):
        """Test repr consistent with object copy"""
        # Not tested with eval due to issues with datetime
        self.out = self.testInst.orbits.__repr__()
        second_out = self.testInst.orbits.copy().__repr__()
        assert self.out == second_out
        return

    def test_load_orbits_w_empty_data(self):
        """ Test orbit loading outside of the instrument data range
        """
        self.stime -= dt.timedelta(days=365 * 100)
        self.testInst.load(date=self.stime)
        self.testInst.orbits[0]
        with pytest.raises(StopIteration):
            self.testInst.orbits.next()

    def test_less_than_one_orbit_of_data(self):
        """Test successful load with less than one orbit of data
        """
        def filter_data(inst):
            """ Local helper function to reduce available data
            """
            inst.data = inst[0:20]

        self.testInst.custom_attach(filter_data)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()

        # a recusion issue has been observed in this area
        # checking for date to limit reintroduction potential
        assert self.testInst.date == self.stime

    def test_less_than_one_orbit_of_data_two_ways(self):
        def filter_data(inst):
            inst.data = inst[0:5]
        self.testInst.custom_attach(filter_data)
        self.testInst.load(date=self.stime)
        # starting from no orbit calls next loads first orbit
        self.testInst.orbits.next()
        # store comparison data
        saved_data = self.testInst.copy()
        self.testInst.load(date=self.stime)
        self.testInst.orbits[0]
        assert all(self.testInst.data == saved_data.data)
        # a recusion issue has been observed in this area
        # checking for date to limit reintroduction potential
        d1check = self.testInst.date == saved_data.date
        assert d1check

    def test_less_than_one_orbit_of_data_four_ways_two_days(self):
        """ Test successful loading of different parital orbits
        """
        # create situation where the < 1 orbit split across two days
        def filter_data(inst):
            """Local function for breaking up orbits
            """
            if inst.date == dt.datetime(2009, 1, 5):
                inst.data = inst[0:20]
            elif inst.date == dt.datetime(2009, 1, 4):
                inst.data = inst[-20:]
            return

        self.testInst.custom_attach(filter_data)
        self.stime += dt.timedelta(days=3)
        self.testInst.load(date=self.stime)

        # starting from no orbit calls next loads first orbit
        self.testInst.orbits.next()

        # store comparison data
        saved_data = self.testInst.copy()
        self.testInst.load(date=self.stime + dt.timedelta(days=1))
        self.testInst.orbits[0]
        if self.testInst.orbits.num == 1:
            # equivalence only when only one orbit
            # some test settings can violate this assumption
            assert all(self.testInst.data == saved_data.data)

        self.testInst.load(date=self.stime)
        self.testInst.orbits[0]
        assert all(self.testInst.data == saved_data.data)

        self.testInst.load(date=self.stime + dt.timedelta(days=1))
        self.testInst.orbits.prev()
        if self.testInst.orbits.num == 1:
            assert all(self.testInst.data == saved_data.data)

        # a recusion issue has been observed in this area
        # checking for date to limit reintroduction potential
        d1check = self.testInst.date == saved_data.date
        assert d1check

    def test_repeated_orbit_calls_symmetric_single_day_start_with_last(self):
        self.testInst.load(date=self.stime)
        # start on last orbit of last day
        self.testInst.orbits[0]
        self.testInst.orbits.prev()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_single_day_0_UT(self):
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_multi_day_0_UT(self):
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_single_day_off_0_UT(self):
        """ Test successful orbit calls for a day about a time off 00:00 UT
        """
        self.stime -= dt.timedelta(days=1)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_multi_day_off_0_UT(self):
        """ Test successful orbit calls for days about a time off 00:00 UT
        """
        self.stime -= dt.timedelta(days=1)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_day_off_0_UT(self):
        """ Test successful orbit calls for different days about a time off 0 UT
        """
        self.stime -= dt.timedelta(days=1)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        for j in range(10):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_off_0_UT(self):
        """ Test successful orbit calls for more days about a time off 0 UT
        """
        self.stime -= dt.timedelta(days=1)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(40):
            self.testInst.orbits.prev()
        for j in range(20):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_day_0_UT(self):
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        for j in range(10):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_0_UT(self):
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(40):
            self.testInst.orbits.prev()
        for j in range(20):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeat_orbit_calls_asym_multi_day_0_UT_long_time_gap(self):
        """Test successful orbit calls for many different days with a long gap
        """
        self.stime += dt.timedelta(days=334)
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeat_orbit_calls_asym_multi_day_0_UT_really_long_time_gap(self):
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(400):
            self.testInst.orbits.next()
        for j in range(400):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeat_orbit_calls_asym_multi_day_0_UT_multiple_time_gaps(self):
        self.testInst.load(date=self.stime)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        n_time = []
        p_time = []
        for j in range(40):
            n_time.append(self.testInst.index[0])
            self.testInst.orbits.next()

        for j in range(40):
            self.testInst.orbits.prev()
            p_time.append(self.testInst.index[0])

        check = np.all(p_time == n_time[::-1])
        assert all(control.data == self.testInst.data) & check


class TestGeneralOrbitsMLTxarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing_xarray._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestGeneralOrbitsNonStandardIteration():
    """Create an iteration window that is larger than step size.
    Ensure the overlapping data doesn't end up in the orbit iteration."""
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'},
                                         update_files=True)
        self.testInst.bounds = (self.testInst.files.files.index[0],
                                self.testInst.files.files.index[11],
                                '2D', dt.timedelta(days=3))
        self.orbit_starts = []
        self.orbit_stops = []

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.orbit_starts, self.orbit_stops

    def test_no_orbit_overlap_with_overlapping_iteration(self):
        """Ensure error when overlap in iteration data."""
        with pytest.raises(ValueError):
            self.testInst.orbits.next()
        return

    @pytest.mark.parametrize("bounds_type", ['by_date', 'by_file'])
    def test_no_orbit_overlap_with_nonoverlapping_iteration(self, bounds_type):
        """Test no orbit data overlap when overlap in iteration data"""

        if bounds_type == 'by_date':
            bounds = (self.testInst.files.files.index[0],
                      self.testInst.files.files.index[11],
                      '2D', dt.timedelta(days=2))
        elif bounds_type == 'by_file':
            bounds = (self.testInst.files[0], self.testInst.files[11], 2, 2)

        self.testInst.bounds = bounds

        for inst in self.testInst.orbits:
            self.orbit_starts.append(inst.index[0])
            self.orbit_stops.append(inst.index[-1])
        self.orbit_starts = pds.Series(self.orbit_starts)
        self.orbit_stops = pds.Series(self.orbit_stops)
        assert self.orbit_starts.is_monotonic_increasing
        assert self.orbit_stops.is_monotonic_increasing
        return


class TestGeneralOrbitsLong(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'longitude',
                                                     'kind': 'longitude'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestGeneralOrbitsLongxarray(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'longitude',
                                                     'kind': 'longitude'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestGeneralOrbitsOrbitNumber(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'orbit_num',
                                                     'kind': 'orbit'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestGeneralOrbitsOrbitNumberXarray(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'orbit_num',
                                                     'kind': 'orbit'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing_xarray._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestGeneralOrbitsLatitude(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'latitude',
                                                     'kind': 'polar'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestGeneralOrbitsLatitudeXarray(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'latitude',
                                                     'kind': 'polar'},
                                         update_files=True)
        self.stime = pysat.instruments.pysat_testing_xarray._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


def filter_data(inst):
    """Remove data from instrument, simulating gaps"""

    times = [[dt.datetime(2009, 1, 1, 1, 37), dt.datetime(2009, 1, 1, 3, 14)],
             [dt.datetime(2009, 1, 1, 10), dt.datetime(2009, 1, 1, 12)],
             [dt.datetime(2009, 1, 1, 22), dt.datetime(2009, 1, 2, 2)],
             [dt.datetime(2009, 1, 13), dt.datetime(2009, 1, 15)],
             [dt.datetime(2009, 1, 20, 1), dt.datetime(2009, 1, 25, 23)],
             [dt.datetime(2009, 1, 25, 23, 30), dt.datetime(2009, 1, 26, 3)]
             ]
    for time in times:
        idx, = np.where((inst.index > time[1]) | (inst.index < time[0]))
        inst.data = inst[idx]


def filter_data2(inst, times=None):
    """Remove data from instrument, simulating gaps"""

    for time in times:
        idx, = np.where((inst.index > time[1]) | (inst.index < time[0]))
        inst.data = inst[idx]


class TestOrbitsGappyData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'},
                                         update_files=True)
        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'},
                                         update_files=True)
        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyData2(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'})
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']
        times = [[dt.datetime(2008, 12, 31, 4),
                  dt.datetime(2008, 12, 31, 5, 37)],
                 [dt.datetime(2009, 1, 1),
                  dt.datetime(2009, 1, 1, 1, 37)]]
        for seconds in np.arange(38):
            day = (dt.datetime(2009, 1, 2)
                   + dt.timedelta(days=int(seconds)))
            times.append([day, day
                          + dt.timedelta(hours=1, minutes=37,
                                         seconds=int(seconds))
                          - dt.timedelta(seconds=20)])

        self.testInst.custom_attach(filter_data2, kwargs={'times': times})

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyData2Xarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'mlt'})
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']
        times = [[dt.datetime(2008, 12, 31, 4),
                  dt.datetime(2008, 12, 31, 5, 37)],
                 [dt.datetime(2009, 1, 1),
                  dt.datetime(2009, 1, 1, 1, 37)]]
        for seconds in np.arange(38):
            day = (dt.datetime(2009, 1, 2)
                   + dt.timedelta(days=int(seconds)))
            times.append([day, day
                          + dt.timedelta(hours=1, minutes=37,
                                         seconds=int(seconds))
                          - dt.timedelta(seconds=20)])

        self.testInst.custom_attach(filter_data2, kwargs={'times': times})

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyLongData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'longitude',
                                                     'kind': 'longitude'})

        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyLongDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'longitude',
                                                     'kind': 'longitude'})
        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyOrbitNumData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'orbit_num',
                                                     'kind': 'orbit'})
        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyOrbitNumDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'orbit_num',
                                                     'kind': 'orbit'})
        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyOrbitLatData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info={'index': 'latitude',
                                                     'kind': 'polar'})

        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime


class TestOrbitsGappyOrbitLatDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info={'index': 'latitude',
                                                     'kind': 'polar'})

        self.testInst.custom_attach(filter_data)
        self.stime = pysat.instruments.pysat_testing._test_dates['']['']

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.stime



"""
pysat - Python Satellite Data Analysis Toolkit
==============================================

pysat is a package providing a simple and flexible interface for
downloading, loading, cleaning, managing, processing, and analyzing
scientific measurements. Although pysat was initially designed for
in situ satellite observations, it now supports many different types
of ground- and space-based measurements.

Main Features
-------------

- Instrument independent analysis routines.
- Instrument object providing an interface for downloading and analyzing
    a wide variety of science data sets.
    - Uses pandas or xarray for the underlying data structure; capable of
        handling the many forms scientific measurements take in a consistent
        manner.
    - Standard scientific data handling tasks (e.g., identifying, downloading,
        and loading files and cleaning and modifying data) are built into the
        Instrument object.
    - Supports metadata consistent with the netCDF CF-1.6 standard. Each
        variable has a name, long name, and units. Note units are informational
        only.
- Simplifies data management
    - Iterator support for loading data by day/file/orbit, independent of
        data storage details.
    - Orbits are calculated on the fly from loaded data and span day breaks.
    - Iterate over custom seasons
- Supports rigorous time-series calculations that require spin up/down
    time across day, orbit, and file breaks.
- Includes helper functions to reduce the barrier in adding new science
    instruments to pysat


"""

import logging
import os
from portalocker import Lock

logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter('%(name)s %(levelname)s: %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.WARNING)

# Import and set user and pysat parameters object
from pysat import _params

# set version
here = os.path.abspath(os.path.dirname(__file__))
version_filename = os.path.join(here, 'version.txt')

# Get home directory
home_dir = os.path.expanduser('~')

# Set pysat directory path in home directory
pysat_dir = os.path.join(home_dir, '.pysat')

# Set directory for test data
test_data_path = os.path.join(here, 'tests', 'test_data')

# Create a .pysat directory or parameters file if one doesn't exist.
# pysat_settings did not exist pre v3 thus this provides a check against
# v2 users that are upgrading. Those users need the settings file plus
# new internal directories.
if not os.path.isdir(pysat_dir) or \
        (not os.path.isfile(os.path.join(pysat_dir, 'pysat_settings.json'))):

    # Make a .pysat directory if not already present
    if not os.path.isdir(pysat_dir):
        os.mkdir(pysat_dir)
        ostr = ''.join(('Created .pysat directory in home directory to store ',
                        'settings.'))
        logger.info(ostr)

    # Make additional internal directories
    if not os.path.isdir(os.path.join(pysat_dir, 'instruments')):
        os.mkdir(os.path.join(pysat_dir, 'instruments'))

    if not os.path.isdir(os.path.join(pysat_dir, 'instruments', 'archive')):
        os.mkdir(os.path.join(pysat_dir, 'instruments', 'archive'))

    # Create parameters file
    if not os.path.isfile(os.path.join(pysat_dir, 'pysat_settings.json')):
        params = _params.Parameters(path=pysat_dir, create_new=True)

    # Set initial data directory if we are on Travis
    if (os.environ.get('TRAVIS') == 'true'):
        data_dir = '/home/travis/build/pysatData'
        params['data_dirs'] = [data_dir]

    print(''.join(("\nHi there!  pysat will nominally store data in a ",
                   "'pysatData' directory which needs to be assigned. ",
                   "Please run `pysat.params['data_dirs'] = path` where path ",
                   "specifies one or more existing top-level directories that ",
                   "may be used to store science data. `path` may either be ",
                   "a single string or a list of strings.")))
else:
    # Load up existing parameters file
    params = _params.Parameters()

# Load up version information
with Lock(version_filename, 'r', params['file_timeout']) as version_file:
    __version__ = version_file.read().strip()

from pysat import utils
from pysat._constellation import Constellation
from pysat._instrument import Instrument
from pysat._meta import Meta, MetaLabels
from pysat._files import Files
from pysat._orbits import Orbits
from pysat import instruments

__all__ = ['instruments', 'utils']

# Cleanup
del here



# -*- coding: utf-8 -*-
"""Supports the Sounding of the Atmosphere using Broadband Emission Radiometry
(SABER) instrument on the Thermosphere Ionosphere Mesosphere Energetics
Dynamics (TIMED) satellite.

Properties
----------
platform
    'timed'
name
    'saber'
tag
    None supported
sat_id
    None supported


Note
----
SABER "Rules of the Road" for DATA USE
Users of SABER data are asked to respect the following guidelines

    - Mission scientific and model results are open to all.
    - Guest investigators, and other members of the scientific community or
      general public should contact the PI or designated team member early in an
      analysis project to discuss the appropriate use of the data.
    - Users that wish to publish the results derived from SABER data should
      normally offer co-authorship to the PI, Associate PI or designated team
      members. Co-authorship may be declined. Appropriate acknowledgement of
      institutions, personnel, and funding agencies should be given.
    - Users should heed the caveats of SABER team members as to the
      interpretation and limitations of the data. SABER team members may insist
      that such caveats be published, even if co-authorship is declined. Data
      and model version numbers should also be specified.
    - Pre-prints of publications and conference abstracts should be widely
      distributed to interested parties within the mission and related projects.


Warnings
--------
- Note on Temperature Errors: http://saber.gats-inc.com/temp_errors.php


Authors
-------
J. Klenzing, 4 March 2019

"""

from __future__ import print_function
from __future__ import absolute_import

import functools

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'timed'
name = 'saber'

# dictionary of data 'tags' and corresponding description
tags = {'': ''}
sat_ids = {'': ['']}

_test_dates = {'': {'': pysat.datetime(2019, 1, 1)}}

fname = ''.join(('timed_l2av207_saber_{year:04d}{month:02d}{day:02d}',
                 '????_v01.cdf'))
supported_tags = {'': {'': fname}}
# use the CDAWeb methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# let pysat know that data is spread across more than one file
multi_file_day = True
pandas_format = True

load = cdw.load
basic_tag = {'dir': '/pub/data/timed/saber/level2a_v2_07_cdf',
             'remote_fname': '{year:4d}/{month:02d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags, multi_file_day=True)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object.

    'clean' All parameters should be good, suitable for statistical and
            case studies
    'dusty' All paramers should generally be good though same may
            not be great
    'dirty' There are data areas that have issues, data should be used
            with caution
    'none'  No cleaning applied, routine not called in this case.


    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    return



# -*- coding: utf-8 -*-
"""Supports the Ion Velocity Meter (IVM)
onboard the Republic of China Satellite (ROCSAT-1). Downloads data from the
NASA Coordinated Data Analysis Web (CDAWeb).

Properties
----------
platform
    'rocsat1'
name
    'ivm'
tag
    None
sat_id
    None supported

Note
----
- no tag or sat_id required

Warnings
--------
- Currently no cleaning routine.

"""

from __future__ import print_function
from __future__ import absolute_import
import functools
import warnings

import pysat

from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'rocsat1'
name = 'ivm'
tags = {'': ''}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(2002, 1, 1)}}


# support list files routine
# use the default CDAWeb method
fname = 'rs_k0_ipei_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)
# support load routine
# use the default CDAWeb method
load = cdw.load

# support download routine
# use the default CDAWeb method
basic_tag = {'dir': '/pub/data/rocsat/ipei',
             'remote_fname': '{year:4d}/'+fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)
# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return ROCSAT-1 IVM data cleaned to the specified level

    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Notes
    --------
    No cleaning currently available for ROCSAT-1 IVM.
    """

    warnings.warn("No cleaning currently available for ROCSAT")

    return None



import datetime as dt
from importlib import reload
import numpy as np
import os

import pysat
from pysat.utils import files as futils
from pysat.tests.travisci_test_class import TravisCICleanSetup


class TestBasics():

    temporary_file_list = False

    def setup(self):
        """Runs before every method to create a clean testing setup."""

    def teardown(self):
        """Runs after every method to clean up previous testing."""

    def test_parse_delimited_filename(self):
        """Check ability to parse list of delimited files"""
        # Note: Can be removed if future instrument that uses delimited
        # filenames is added to routine travis end-to-end testing
        fname = ''.join(('test_{year:4d}_{month:2d}_{day:2d}_{hour:2d}',
                         '_{minute:2d}_{second:2d}_{version:2s}_r02.cdf'))
        year = np.ones(6) * 2009
        month = np.ones(6) * 12
        day = np.array([12, 15, 17, 19, 22, 24])
        hour = np.array([8, 10, 6, 18, 3, 23])
        minute = np.array([8, 10, 6, 18, 3, 59])
        second = np.array([58, 11, 26, 2, 18, 59])
        version = np.array(['v1', 'v2', 'r1', 'r3', 'v5', 'a6'])
        file_list = []
        for i in range(6):
            file_list.append(fname.format(year=year[i].astype(int),
                                          month=month[i].astype(int),
                                          day=day[i], hour=hour[i],
                                          minute=minute[i], second=second[i],
                                          version=version[i]))

        file_dict = futils.parse_delimited_filenames(file_list, fname, '_')
        assert np.all(file_dict['year'] == year)
        assert np.all(file_dict['month'] == month)
        assert np.all(file_dict['day'] == day)
        assert np.all(file_dict['hour'] == hour)
        assert np.all(file_dict['minute'] == minute)
        assert np.all(file_dict['day'] == day)
        assert np.all(file_dict['version'] == version)
        assert (file_dict['revision'] is None)
        assert (file_dict['cycle'] is None)


class TestFileDirectoryTranslations(TravisCICleanSetup):

    def setup(self):
        """Runs before every method to create a clean testing setup."""

        # Module is only required for testing installations on TravisCI
        import pysatSpaceWeather

        # Create clean environment on Travis
        TravisCICleanSetup.setup(self)
        reload(pysat)

        # Note, if testing locally, after setting self.ci_env = True
        # in TravisCICleanSetup.setup then a data directory needs to be
        # set here.
        # pysat.params['data_dirs'] = '~/DemoData/'

        # Create several pysat.SpaceWeather instruments and download data.
        # We want to start with a setup that covers general cases a user may
        # have.
        pysat.utils.registry.register_by_module(pysatSpaceWeather.instruments)

        self.insts = []
        self.insts_dates = []
        self.insts_kwargs = []

        # Data by day, ACE SIS data
        self.insts.append(pysat.Instrument('sw', 'ace', tag='historic',
                                           inst_id='sis'))
        test_dates = pysatSpaceWeather.instruments.sw_ace._test_dates
        self.insts_dates.append([test_dates['sis']['historic']] * 2)
        self.insts_kwargs.append({})

        # Data with date mangling, regular F10.7 data, stored monthly
        self.insts.append(pysat.Instrument('sw', 'f107'))
        test_dates = pysatSpaceWeather.instruments.sw_f107._test_dates
        self.insts_dates.append([test_dates[''][''],
                                 test_dates[''][''] + dt.timedelta(weeks=52)])
        self.insts_kwargs.append({'freq': 'MS'})

        # Data with date mangling, 'all' F10.7 data, single file
        self.insts.append(pysat.Instrument('sw', 'f107', tag='all'))
        test_dates = pysatSpaceWeather.instruments.sw_f107._test_dates
        self.insts_dates.append([test_dates['']['all'],
                                 test_dates['']['all']])
        self.insts_kwargs.append({})

        # Download data for all instruments
        for inst, dates, kwargs in zip(self.insts, self.insts_dates,
                                       self.insts_kwargs):
            ostr = ' '.join(('Downloading data for', inst.platform,
                             inst.name, inst.tag, inst.inst_id))
            print(ostr)
            inst.download(start=dates[0], stop=dates[1], **kwargs)

    def teardown(self):
        """Runs after every method to clean up previous testing."""

        # Clean environment on Travis
        TravisCICleanSetup.teardown(self)

        # TODO: Check on potentially removing all inst directories

    def test_updating_directories(self, capsys):
        """Test directory structure update method"""

        # A variety of options are performed within a single test
        # so that only one round of downloads is performed.

        # New Template
        templ = '{platform}'

        # Convert directories to simpler platform structure, as a test run,
        # with expanded descriptive output.
        futils.update_data_directory_structure(new_template=templ,
                                               full_breakdown=True)

        # Capture printouts
        captured = capsys.readouterr()

        # Check for descriptive output from full_breakdown
        fstr = 'Will move: '
        assert captured.out.find(fstr) >= 0

        # Check how many instruments have no files found. Will be used later.
        index = 0
        orig_num_missing = 0
        while index < len(captured.out):
            index = captured.out.find('No files found.', index)
            if index == -1:
                break
            else:
                index += 1
                orig_num_missing += 1

        # Convert directories to simpler platform structure
        futils.update_data_directory_structure(new_template=templ,
                                               test_run=False,
                                               remove_empty_dirs=True,
                                               full_breakdown=True)

        # Capture printouts
        captured = capsys.readouterr()

        # Check if we did things correctly. Look for correct output strings.
        for inst in self.insts:
            # Check for all files moved.
            fstr = ' '.join(('All', inst.platform, inst.name, inst.tag,
                             inst.inst_id, 'files moved and accounted for.',
                             '\n'))
            assert captured.out.find(fstr) >= 0

            # Check that directories were removed.
            assert not os.path.isdir(inst.files.data_path)

        # Store newly organized directory format
        pysat.params['directory_format'] = templ

        # Convert directories back to more complex structure
        # First, define new template
        templ = '{platform}/{name}/{tag}/{inst_id}'

        # Update structure
        futils.update_data_directory_structure(new_template=templ,
                                               test_run=False,
                                               remove_empty_dirs=True,
                                               full_breakdown=True)

        # Capture printouts
        captured = capsys.readouterr()

        # Check if we did things correctly. Look for correct output strings.
        for inst in self.insts:
            # Check for all files moved.
            fstr = ' '.join(('All', inst.platform, inst.name, inst.tag,
                             inst.inst_id, 'files moved and accounted for.',
                             '\n'))
            assert captured.out.find(fstr) >= 0

            # Refresh inst with the old directory template set to get now 'old'
            # path information.
            inst2 = pysat.Instrument(inst.platform, inst.name, tag=inst.tag,
                                     inst_id=inst.inst_id)

            # Check that directories with simpler platform org were NOT removed.
            assert os.path.isdir(inst2.files.data_path)

            # Confirm that the code was aware the directory was not empty.
            fstr = ''.join(('Directory is not empty: ',
                            inst2.files.data_path, '\nEnding cleanup.', '\n'))
            assert captured.out.find(fstr) >= 0

        # Try to update structure again. Files have already moved so
        # no files should be found.
        futils.update_data_directory_structure(new_template=templ,
                                               test_run=False,
                                               remove_empty_dirs=True,
                                               full_breakdown=True)

        # Capture printouts
        captured = capsys.readouterr()

        # Check for no files output
        index = 0
        num_missing = 0
        while index < len(captured.out):
            index = captured.out.find('No files found.', index)
            if index == -1:
                break
            else:
                index += 1
                num_missing += 1

        # Get difference in number of instruments with no files.
        new_missing = num_missing - orig_num_missing

        # Confirm none of the instruments had files.
        assert new_missing == len(self.insts)

        # Store new format like a typical user would
        pysat.params['directory_format'] = templ



from dateutil.relativedelta import relativedelta as relativedelta
from nose.tools import raises
import numpy as np
import pandas as pds

import pysat


class TestOrbitsUserInterface():

    @raises(ValueError)
    def test_orbit_w_bad_kind(self):
        info = {'index': 'mlt', 'kind': 'cats'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)        

    @raises(ValueError)
    def test_orbit_long_w_bad_variable(self):
        info = {'index': 'magnetic local time', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_long_w_missing_orbit_info(self):
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         update_files=True)
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_long_w_missing_orbit_index(self):
        info = {'index': 'magnetic local time', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        # force index to None
        self.testInst.orbits.orbit_index = None
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_mlt_w_bad_variable(self):
        info = {'index': 'magnetic local time', 'kind': 'lt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_mlt_w_missing_orbit_index(self):
        info = {'index': 'magnetic local time', 'kind': 'lt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        # force index to None
        self.testInst.orbits.orbit_index = None
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_polar_w_bad_variable(self):
        info = {'index': 'magnetic local time', 'kind': 'polar'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_polar_w_missing_orbit_index(self):
        info = {'index': 'magnetic local time', 'kind': 'polar'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        # force index to None
        self.testInst.orbits.orbit_index = None
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_orbibt_w_bad_variable(self):
        info = {'index': 'magnetic local time', 'kind': 'orbit'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()

    @raises(ValueError)
    def test_orbit_orbit_w_missing_orbit_index(self):
        info = {'index': 'magnetic local time', 'kind': 'orbit'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        # force index to None
        self.testInst.orbits.orbit_index = None
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        

class TestSpecificUTOrbits():

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    def test_single_orbit_call_by_0_index(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits[0]
        ans = (self.testInst.index[0] == pds.datetime(2009, 1, 1))
        ans2 = (self.testInst.index[-1] == pds.datetime(2009, 1, 1, 1, 36, 59))
        assert ans & ans2

    def test_single_orbit_call_by_1_index(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits[1]
        ans = (self.testInst.index[0] == pds.datetime(2009, 1, 1, 1, 37))
        ans2 = (self.testInst.index[-1] == pds.datetime(2009, 1, 1, 3, 13, 59))
        assert ans & ans2

    def test_single_orbit_call_by_negative_1_index(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits[-1]
        ans = (self.testInst.index[0] ==
               (pds.datetime(2009, 1, 1)-relativedelta(hours=1, minutes=37)))
        ans2 = (self.testInst.index[-1] ==
                (pds.datetime(2009, 1, 1)-relativedelta(seconds=1)))
        assert ans & ans2

    def test_single_orbit_call_by_last_index(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits[14]
        assert (self.testInst.index[0] ==
                (pds.datetime(2009, 1, 1)-relativedelta(hours=1, minutes=37)))
        assert (self.testInst.index[-1] ==
                (pds.datetime(2009, 1, 1)-relativedelta(seconds=1)))

    @raises(Exception)
    def test_single_orbit_call_too_many(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits[17]

    @raises(Exception)
    def test_single_orbit_no_input(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits[None]

    def test_oribt_number_via_current_multiple_orbit_calls_in_day(self):
        self.testInst.load(2009, 1)
        self.testInst.bounds = (pysat.datetime(2009, 1, 1), None)
        true_vals = np.arange(15)
        true_vals[-1] = 0
        test_vals = []
        for i, inst in enumerate(self.testInst.orbits):
            if i > 14:
                break
            test_vals.append(self.testInst.orbits.current)
            print('Loaded orbit ', self.testInst.orbits.current)
 
        assert np.all(test_vals == true_vals)

    def test_all_single_orbit_calls_in_day(self):
        self.testInst.load(2009, 1)
        ans = []
        ans2 = []
        self.testInst.bounds = (pysat.datetime(2009, 1, 1), None)
        for i, inst in enumerate(self.testInst.orbits):
            if i > 14:
                break
            print('Loaded orbit ', self.testInst.orbits.current)
            ans.append(self.testInst.index[0] ==
                       (pds.datetime(2009, 1, 1) +
                       i*relativedelta(hours=1, minutes=37)))
            ans2.append(self.testInst.index[-1] ==
                        (pds.datetime(2009, 1, 1) +
                        (i + 1) * relativedelta(hours=1, minutes=37) -
                        relativedelta(seconds=1)))

        assert np.all(ans) & np.all(ans2)

    def test_orbit_next_call_no_loaded_data(self):
        self.testInst.orbits.next()
        assert (self.testInst.index[0] == pds.datetime(2008, 1, 1))
        assert (self.testInst.index[-1] == pds.datetime(2008, 1, 1, 0, 38, 59))

    def test_orbit_prev_call_no_loaded_data(self):
        self.testInst.orbits.prev()
        # this isn't a full orbit
        assert (self.testInst.index[-1] ==
                pds.datetime(2010, 12, 31, 23, 59, 59))
        assert (self.testInst.index[0] == pds.datetime(2010, 12, 31, 23, 49))

    def test_single_orbit_call_orbit_starts_0_UT_using_next(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        assert (self.testInst.index[0] == pds.datetime(2009, 1, 1))
        assert (self.testInst.index[-1] == pds.datetime(2009, 1, 1, 1, 36, 59))

    def test_single_orbit_call_orbit_starts_0_UT_using_prev(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.prev()
        assert (self.testInst.index[0] ==
                (pds.datetime(2009, 1, 1) +
                14 * relativedelta(hours=1, minutes=37)))
        assert (self.testInst.index[-1] ==
                (pds.datetime(2009, 1, 1) +
                15 * relativedelta(hours=1, minutes=37) -
                relativedelta(seconds=1)))

    def test_single_orbit_call_orbit_starts_off_0_UT_using_next(self):
        from dateutil.relativedelta import relativedelta as relativedelta
        self.testInst.load(2008, 366)
        self.testInst.orbits.next()
        assert (self.testInst.index[0] == pds.datetime(2008, 12, 30, 23, 45))
        assert (self.testInst.index[-1] ==
                (pds.datetime(2008, 12, 30, 23, 45) +
                relativedelta(hours=1, minutes=36, seconds=59)))

    def test_single_orbit_call_orbit_starts_off_0_UT_using_prev(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits.prev()
        assert (self.testInst.index[0] ==
                (pds.datetime(2009, 1, 1)-relativedelta(hours=1, minutes=37)))
        assert (self.testInst.index[-1] ==
                (pds.datetime(2009, 1, 1)-relativedelta(seconds=1)))


class TestGeneralOrbitsMLT():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    @raises(StopIteration)
    def test_load_orbits_w_empty_data(self):
        self.testInst.load(1958, 31)
        self.testInst.orbits[0]
        self.testInst.orbits.next()

    def test_less_than_one_orbit_of_data(self):
        def filter_data(inst):
            inst.data = inst[0:20]
        self.testInst.custom.add(filter_data, 'modify')
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        # a recusion issue has been observed in this area
        # checking for date to limit reintroduction potential
        assert self.testInst.date == pysat.datetime(2009, 1, 1)

    def test_less_than_one_orbit_of_data_two_ways(self):
        def filter_data(inst):
            inst.data = inst[0:5]
        self.testInst.custom.add(filter_data, 'modify')
        self.testInst.load(2009, 1)
        # starting from no orbit calls next loads first orbit
        self.testInst.orbits.next()
        # store comparison data
        saved_data = self.testInst.copy()
        self.testInst.load(2009, 1)
        self.testInst.orbits[0]
        assert all(self.testInst.data == saved_data.data)
        # a recusion issue has been observed in this area
        # checking for date to limit reintroduction potential
        d1check = self.testInst.date == saved_data.date
        assert d1check 

    def test_less_than_one_orbit_of_data_four_ways_two_days(self):
        # create situation where the < 1 orbit split across two days
        def filter_data(inst):
            if inst.date == pysat.datetime(2009, 1, 5):
                inst.data = inst[0:20]
            elif inst.date == pysat.datetime(2009, 1, 4):
                inst.data = inst[-20:]

        self.testInst.custom.add(filter_data, 'modify')
        self.testInst.load(2009, 4)
        # starting from no orbit calls next loads first orbit
        self.testInst.orbits.next()
        # store comparison data
        saved_data = self.testInst.copy()
        self.testInst.load(2009, 5)
        self.testInst.orbits[0]
        if self.testInst.orbits.num == 1:
            # equivalence only when only one orbit
            # some test settings can violate this assumption
            assert all(self.testInst.data == saved_data.data)
        else:
            print('Skipping this part of test.')
        self.testInst.load(2009, 4)
        self.testInst.orbits[0]
        assert all(self.testInst.data == saved_data.data)
        
        self.testInst.load(2009, 5)
        self.testInst.orbits.prev()
        if self.testInst.orbits.num == 1:
            assert all(self.testInst.data == saved_data.data)
        else:
            print('Skipping this part of test.')
        # a recusion issue has been observed in this area
        # checking for date to limit reintroduction potential
        d1check = self.testInst.date == saved_data.date
        assert d1check

    def test_repeated_orbit_calls_symmetric_single_day_start_with_last(self):
        self.testInst.load(2009, 1)
        # start on last orbit of last day
        self.testInst.orbits[0]
        self.testInst.orbits.prev()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_single_day_0_UT(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_multi_day_0_UT(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_single_day_off_0_UT(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(10):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_symmetric_multi_day_off_0_UT(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_day_off_0_UT(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        for j in range(10):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_off_0_UT(self):
        self.testInst.load(2008, 366)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(40):
            self.testInst.orbits.prev()
        for j in range(20):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_day_0_UT(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(10):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        for j in range(10):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeated_orbit_calls_antisymmetric_multi_multi_day_0_UT(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(40):
            self.testInst.orbits.prev()
        for j in range(20):
            self.testInst.orbits.next()
        assert all(control.data == self.testInst.data)

    def test_repeat_orbit_calls_asym_multi_day_0_UT_long_time_gap(self):
        self.testInst.load(2009, 12)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(20):
            self.testInst.orbits.next()
        for j in range(20):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeat_orbit_calls_asym_multi_day_0_UT_really_long_time_gap(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        for j in range(400):
            self.testInst.orbits.next()
        for j in range(400):
            self.testInst.orbits.prev()
        assert all(control.data == self.testInst.data)

    def test_repeat_orbit_calls_asym_multi_day_0_UT_multiple_time_gaps(self):
        self.testInst.load(2009, 1)
        self.testInst.orbits.next()
        control = self.testInst.copy()
        n_time = []
        p_time = []
        for j in range(40):
            n_time.append(self.testInst.index[0])
            self.testInst.orbits.next()

        for j in range(40):
            self.testInst.orbits.prev()
            p_time.append(self.testInst.index[0])

        check = np.all(p_time == n_time[::-1])
        assert all(control.data == self.testInst.data) & check


class TestGeneralOrbitsMLTxarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestGeneralOrbitsLong(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'longitude', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestGeneralOrbitsLongxarray(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'longitude', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestGeneralOrbitsOrbitNumber(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'orbit_num', 'kind': 'orbit'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestGeneralOrbitsOrbitNumberXarray(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'orbit_num', 'kind': 'orbit'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestGeneralOrbitsLatitude(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'latitude', 'kind': 'polar'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestGeneralOrbitsLatitudeXarray(TestGeneralOrbitsMLT):

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'latitude', 'kind': 'polar'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


def filter_data(inst):
    """Remove data from instrument, simulating gaps"""

    times = [[pysat.datetime(2009, 1, 1, 10),
              pysat.datetime(2009, 1, 1, 12)],
             [pysat.datetime(2009, 1, 1, 4),
              pysat.datetime(2009, 1, 2, 5, 37)],
             [pysat.datetime(2009, 1, 1, 1, 37),
              pysat.datetime(2009, 1, 1, 3, 14)],
             [pysat.datetime(2009, 1, 1, 15),
              pysat.datetime(2009, 1, 1, 16)],
             [pysat.datetime(2009, 1, 1, 22),
              pysat.datetime(2009, 1, 2, 2)],
             [pysat.datetime(2009, 1, 13),
              pysat.datetime(2009, 1, 15)],
             [pysat.datetime(2009, 1, 20, 1),
              pysat.datetime(2009, 1, 25, 23)],
             [pysat.datetime(2009, 1, 25, 23, 30),
              pysat.datetime(2009, 1, 26, 3)]
             ]
    for time in times:
        idx, = np.where((inst.index > time[1]) | (inst.index < time[0]))
        inst.data = inst[idx]


def filter_data2(inst, times=None):
    """Remove data from instrument, simulating gaps"""

    for time in times:
        idx, = np.where((inst.index > time[1]) | (inst.index < time[0]))
        inst.data = inst[idx]


class TestOrbitsGappyData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info, update_files=True)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""


class TestOrbitsGappyData2(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info)
        times = [[pysat.datetime(2008, 12, 31, 4),
                  pysat.datetime(2008, 12, 31, 5, 37)],
                 [pysat.datetime(2009, 1, 1),
                  pysat.datetime(2009, 1, 1, 1, 37)]
                 ]
        for seconds in np.arange(38):
            day = pysat.datetime(2009, 1, 2) + \
                pds.DateOffset(days=int(seconds))
            times.append([day, day +
                          pds.DateOffset(hours=1, minutes=37,
                                         seconds=int(seconds)) -
                          pds.DateOffset(seconds=20)])

        self.testInst.custom.add(filter_data2, 'modify', times=times)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyData2Xarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'mlt'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info)
        times = [[pysat.datetime(2008, 12, 31, 4),
                  pysat.datetime(2008, 12, 31, 5, 37)],
                 [pysat.datetime(2009, 1, 1),
                  pysat.datetime(2009, 1, 1, 1, 37)]
                 ]
        for seconds in np.arange(38):
            day = pysat.datetime(2009, 1, 2) + \
                pds.DateOffset(days=int(seconds))
            times.append([day, day +
                          pds.DateOffset(hours=1, minutes=37,
                                         seconds=int(seconds)) -
                          pds.DateOffset(seconds=20)])

        self.testInst.custom.add(filter_data2, 'modify', times=times)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyLongData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'longitude', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyLongDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'longitude', 'kind': 'longitude'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyOrbitNumData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'orbit_num', 'kind': 'orbit'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyOrbitNumDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'orbit_num', 'kind': 'orbit'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyOrbitLatData(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'latitude', 'kind': 'polar'}
        self.testInst = pysat.Instrument('pysat', 'testing',
                                         clean_level='clean',
                                         orbit_info=info)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class TestOrbitsGappyOrbitLatDataXarray(TestGeneralOrbitsMLT):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        info = {'index': 'latitude', 'kind': 'polar'}
        self.testInst = pysat.Instrument('pysat', 'testing_xarray',
                                         clean_level='clean',
                                         orbit_info=info)
        self.testInst.custom.add(filter_data, 'modify')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst



# -*- coding: utf-8 -*-
"""Supports the Planar Langmuir Probe (PLP) onboard the Communication
and Navigation Outage Forecasting System (C/NOFS) satellite. Downloads
data from the NASA Coordinated Data Analysis Web (CDAWeb).

Description from CDAWeb:

The Planar Langmuir Probe on C/NOFS is a suite of 2 current measuring sensors
mounted on the ram facing surface of the spacecraft.  The primary sensor is an
Ion Trap (conceptually similar to RPAs flown on many other spacecraft) capable
of measuring ion densities as low as 1 cm-3 with a 12 bit log electrometer.
The secondary senor is a swept bias planar Langmuir probe (Surface Probe)
capable of measuring Ne, Te, and spacecraft potential.

The ion number density is the one second average of the ion density sampled at
either 32, 256, 512, or 1024 Hz (depending on the mode).

The ion density standard deviation is the standard deviation of the samples
used to produce the one second average number density.

DeltaN/N is the detrened ion number density 1 second standard deviation divided
by the mean 1 sec density.

The electron density, electron temperature, and spacecraft potential are all
derived from a least squares fit to the current-bias curve from the Surface
Probe.

The data is PRELIMINARY, and as such, is intended for BROWSE PURPOSES ONLY.

References
----------
A brief discussion of the C/NOFS mission and instruments can be found at
de La Beaujardière, O., et al. (2004), C/NOFS: A mission to forecast
scintillations, J. Atmos. Sol. Terr. Phys., 66, 1573–1591,
doi:10.1016/j.jastp.2004.07.030.

Properties
----------
platform
    'cnofs'
name
    'plp'
tag
    None supported
sat_id
    None supported

Warnings
--------
- The data are PRELIMINARY, and as such, are intended for BROWSE PURPOSES ONLY.
- Currently no cleaning routine.
- Module not written by PLP team.

"""

from __future__ import print_function
from __future__ import absolute_import
import functools
import numpy as np

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

platform = 'cnofs'
name = 'plp'
tags = {'': ''}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(2009, 1, 1)}}


# support list files routine
# use the default CDAWeb method
fname = 'cnofs_plp_plasma_1sec_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)
# support load routine
# use the default CDAWeb method
load = cdw.load

# support download routine
# use the default CDAWeb method
basic_tag = {'dir': '/pub/data/cnofs/plp/plasma_1sec',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)
# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(inst):
    """Routine to return C/NOFS PLP data cleaned to the specified level

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Notes
    --------
    Basic cleaning to find valid Epoch values

    """

    for key in inst.data.columns:
        if key != 'Epoch':
            idx, = np.where(inst[key] == inst.meta[key, inst.fill_label])
            inst[idx, key] = np.nan



import requests
import warnings

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw


class TestCDAWeb():

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.supported_tags = pysat.instruments.cnofs_plp.supported_tags

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.supported_tags

    def test_remote_file_list_connection_error_append(self):
        """Test that pysat appends suggested help to ConnectionError"""
        # Giving a bad remote_site address yields similar ConnectionError
        try:
            cdw.list_remote_files(tag='', sat_id='',
                                  supported_tags=self.supported_tags,
                                  remote_site='http:/')
        except Exception as excinfo:
            assert isinstance(excinfo, requests.exceptions.ConnectionError)
            # Check that pysat appends the message
            assert str(excinfo).find('pysat -> Request potentially') > 0
        else:
            raise(ValueError, 'Test was expected to raise an Exception.')

    def test_remote_file_list_deprecation_warning(self):
        """Test generation of deprecation warning for remote_file_list kwargs
        """
        warnings.simplefilter("always")

        with warnings.catch_warnings(record=True) as war:
            # testing with single day since we just need the warning
            cdw.list_remote_files(tag='', sat_id='',
                                  supported_tags=self.supported_tags,
                                  year=2009, month=1, day=1)

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_list_files_deprecation_warning(self):
        """Test generation of deprecation warning for list_files kwargs
        """
        warnings.simplefilter("always")

        with warnings.catch_warnings(record=True) as war1:
            # testing with single day since we just need the warning
            try:
                pysat.instruments.methods.nasa_cdaweb.list_files()
            except ValueError:
                # Using default tags will produce a ValueError
                pass

        with warnings.catch_warnings(record=True) as war2:
            # testing with single day since we just need the warning
            try:
                pysat.instruments.methods.general.list_files()
            except ValueError:
                # Using default tags will produce a ValueError
                pass

        assert len(war1) >= 1
        assert war1[0].category == DeprecationWarning
        assert len(war2) == 0



"""
pysat - Python Satellite Data Analysis Toolkit
==============================================

pysat is a package providing a simple and flexible interface for
downloading, loading, cleaning, managing, processing, and analyzing
scientific measurements. Although pysat was initially designed for
in situ satellite observations, it now supports many different types
of ground- and space-based measurements.

Main Features
-------------

- Instrument independent analysis routines.
- Instrument object providing an interface for downloading and analyzing
    a wide variety of science data sets.
    - Uses pandas or xarray for the underlying data structure; capable of
        handling the many forms scientific measurements take in a consistent
        manner.
    - Standard scientific data handling tasks (e.g., identifying, downloading,
        and loading files and cleaning and modifying data) are built into the
        Instrument object.
    - Supports metadata consistent with the netCDF CF-1.6 standard. Each
        variable has a name, long name, and units. Note units are informational
        only.
- Simplifies data management
    - Iterator support for loading data by day/file/orbit, independent of
        data storage details.
    - Orbits are calculated on the fly from loaded data and span day breaks.
    - Iterate over custom seasons
- Supports rigorous time-series calculations that require spin up/down
    time across day, orbit, and file breaks.
- Includes helper functions to reduce the barrier in adding new science
    instruments to pysat


"""

# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
import os

import logging
logger = logging.getLogger(__name__)
handler = logging.StreamHandler()
formatter = logging.Formatter('%(name)s %(levelname)s: %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)
logger.setLevel(logging.WARNING)

# set version
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'version.txt')) as version_file:
    __version__ = version_file.read().strip()


# get home directory
home_dir = os.path.expanduser('~')
# Set directory for test data
test_data_path = os.path.join(here, 'tests', 'test_data')
# set pysat directory path in home directory
pysat_dir = os.path.join(home_dir, '.pysat')
# make sure a pysat directory exists
if not os.path.isdir(pysat_dir):
    # create directory
    os.mkdir(pysat_dir)
    print('Created .pysat directory in user home directory to store settings.')
    # create file with default data directory
    if (os.environ.get('TRAVIS') == 'true'):
        data_dir = '/home/travis/build/pysatData'
    else:
        data_dir = ''
    with open(os.path.join(pysat_dir, 'data_path.txt'), 'w') as f:
        f.write(data_dir)
    print(''.join(("\nHi there!  Please inform pysat where you will store "
                   "(or are storing) science data by "
                   "running pysat.utils.set_data_dir and specifying "
                   "a location.")))
else:
    # load up stored data path
    with open(os.path.join(pysat_dir, 'data_path.txt'), 'r') as f:
        data_dir = f.readline()

import netCDF4
from pandas import Panel, DataFrame, Series, datetime
from . import utils, model_utils
from ._constellation import Constellation
from ._instrument import Instrument
from ._meta import Meta
from ._files import Files
from ._custom import Custom
from ._orbits import Orbits
from . import instruments
from . import ssnl

__all__ = ['ssnl', 'instruments', 'utils']



def computational_form(data):
    """
    Repackages numbers, Series, or DataFrames

    .. deprecated:: 2.2.0
      `computational_form` will be removed in pysat 3.0.0, it will
      be added to pysatSeasons

    Regardless of input format, mathematical operations may be performed on the
    output via the same pandas mechanisms.

    This method may be particularly useful in analysis methods that aim to be
    instrument independent. pysat.Instrument objects can package data in a
    variety of ways within a DataFrame, depending upon the scientific data
    source. Thus, a variety of data types will be encountered by instrument
    independent methods and computational_form method may reduce the effort
    required to support more generalized processing.

    Parameters
    ----------
    data : pandas.Series
        Series of numbers, Series, DataFrames

    Returns
    -------
    pandas.Series, DataFrame, or Panel
        repacked data, aligned by indices, ready for calculation
    """

    from pysat import DataFrame, Series, Panel
    import warnings

    warnings.warn(' '.join(["This function is deprecated here and will be",
                            "removed in pysat 3.0.0. Please use",
                            "pysatSeasons instead:"
                            "https://github.com/pysat/pysatSeasons"]),
                  DeprecationWarning, stacklevel=2)

    if isinstance(data.iloc[0], DataFrame):
        dslice = Panel.from_dict(dict([(i, data.iloc[i])
                                       for i in range(len(data))]))
    elif isinstance(data.iloc[0], Series):
        dslice = DataFrame(data.tolist())
        dslice.index = data.index
    else:
        dslice = data
    return dslice



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

import copy
import json
import numpy as np
import os

from portalocker import Lock

from pysat.utils.files import check_and_make_path


class Parameters(object):
    """Stores user parameters used by pysat.

    Also stores custom user parameters provided the keys don't conflict with
    default pysat parameters.

    Parameters
    ----------
    path : str
        If provided, the directory path will be used to load/store a
        parameters file with name 'pysat_settings.json' (default=None).
    create_new : bool
        If True, a new parameters file is created. Will be created at path
        if provided. If not, file will be created in `pysat.pysat_dir`.

    Attributes
    ----------
    data : dict
        pysat user settings dictionary
    defaults : dict
        Default parameters (keys) and values used by pysat that include
        {'clean_level': 'clean', 'directory_format':
        os.path.join('{platform}', '{name}', '{tag}', '{inst_id}'),
        'ignore_empty_files': False, 'update_files': True,
        'file_timeout': 10, 'user_modules' : {}, 'warn_empty_file_list': False}
    file_path : str
        Location of file used to store settings
    non_defaults : list
        List of pysat parameters (strings) that don't have a defined
        default and are unaffected by self.restore_defaults().

    Raises
    ------
    ValueError
        The 'user_modules' parameter may not be set directly by the
        user. Please use the `pysat.utils.regsitry` module to modify
        the packages stored in 'user_modules'.

    Note
    ----
    This method will look for 'pysat_settings.json' file first in the
    current working directory and then in the home '~/.pysat' directory.

    All pysat parameters are automatically stored whenever a parameter is
    assigned or modified. The default parameters and values tracked by this
    class are grouped by type below.

    Values that map to the corresponding keywords on pysat.Instrument:
    clean_level, directory_format, ignore_empty_files, and update_files.  See
    the Instrument docstring for more information on these keywords.

    Values that map to internal pysat settings: file_timeout, user_modules, and
    warn_empty_file_list.

    Stored pysat parameters without a working default value: data_dirs.

    file_timeout -  Time in seconds that pysat will wait to modify a busy file

    user_modules - Stores information on modules registered by pysat

    warn_empty_file_list - Raise a warning when no Instrument files are found

    data_dirs - Directory(ies) where data are stored, in access order

    """

    def __init__(self, path=None, create_new=False):
        sfname = 'pysat_settings.json'
        self.data = {}
        self.file_path = None

        # Define default parameters and values
        dir_format = os.path.join('{platform}', '{name}', '{tag}', '{inst_id}')
        defaults = {'clean_level': 'clean',
                    'directory_format': dir_format,
                    'ignore_empty_files': False,
                    'file_timeout': 10,
                    'update_files': True,
                    'user_modules': {},
                    'warn_empty_file_list': False}

        # Attach default parameters and values to object
        self.defaults = defaults

        # Define stored pysat parameters without a default setting
        non_defaults = ['data_dirs']
        self.non_defaults = non_defaults

        # If path provided, use it. Otherwise, iterate through potential
        # locations until file is found.
        if path is not None:
            # Confirm path is valid
            if not os.path.exists(path):
                estr = ''.join(('Supplied path does not exist on the local ',
                                'system. Please create it and try again.'))
                raise OSError(estr)

            # Store full file path including fixed settings file name
            self.file_path = os.path.join(path, sfname)

        else:
            # Cycle through locations and load any pysat parameter files found.
            # First, check current working directory, then pysat user directory.
            # Stop looking once an existing file is found.
            file_locs = [os.path.join('.', sfname),
                         os.path.join(os.path.expanduser('~'), '.pysat',
                                      sfname)]
            for fileloc in file_locs:
                if os.path.isfile(fileloc):
                    self.file_path = fileloc
                    break

            # Ensure we have a valid file if the user isn't creating a new one.
            if self.file_path is None and (not create_new):
                estr = ''.join(('pysat is unable to locate a user settings ',
                                'file. Please check the locations, "./" or ',
                                '"~/.pysat" for the file "pysat_settings.json"',
                                '.'))
                raise OSError(estr)

        if create_new:
            # Initialize new settings file. Method below includes a .store call
            self.clear_and_restart()

        # Load parameters in thread-safe manner.
        # Can't use user set file_timeout since we don't know what it is yet.
        # Can't use NetworkLock in pysat.utils either since this object (params)
        # doesn't exist yet.
        with Lock(self.file_path, 'r', timeout=10) as fout:
            self.data = json.load(fout)
            # In case of network file system
            fout.flush()
            os.fsync(fout.fileno())

        return

    def __repr__(self):
        """String describing Parameters instantiation parameters

        Returns
        -------
        out_str : str
            Simply formatted output string

        """
        dir_path = os.path.split(self.file_path)[0]
        out_str = ''.join(('pysat._params.Parameters(path="', dir_path, '")'))
        return out_str

    def __str__(self, long_str=True):
        """String describing Parameters instance, variables, and attributes

        Parameters
        ----------
        long_str : bool
            Return short version if False and long version if True
            (default=True)

        Returns
        -------
        out_str : str
            Nicely formatted output string

        """

        # Get typical pysat parameters (those with defaults)
        typical = list(self.defaults.keys())

        # Get pysat parameters without working defaults
        pparams = self.non_defaults

        # Get any additional parameters set by the user
        both = typical.copy()
        both.extend(pparams)
        users = [key for key in self.data.keys() if key not in both]

        # Print the short output
        out_str = "pysat Parameters object\n"
        out_str += "----------------------\n"
        out_str += "Tracking {:d} pysat settings\n".format(len(typical))
        out_str += "Tracking {:d} settings (non-default)\n".format(len(pparams))
        out_str += "Tracking {:d} user values\n".format(len(users))

        # Print the longer output
        if long_str:

            out_str += "\nStandard parameters:\n"
            for typ in typical:
                out_str += ''.join((typ, ' : ', self[typ].__str__(), '\n'))

            out_str += "\nStandard parameters (no defaults):\n"
            for typ in pparams:
                out_str += ''.join((typ, ' : ', self[typ].__str__(), '\n'))

            out_str += "\nUser parameters:\n"
            for typ in users:
                out_str += ''.join((typ, ' : ', self[typ].__str__(), '\n'))

        return out_str

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        # Update current settings
        # Some parameters require processing before storage.
        if key == 'data_dirs':
            self._set_data_dirs(value)

        elif key == 'user_modules':
            estr = ''.join(('The pysat.utils.registry ',
                            'submodule has methods designed to build ',
                            'and work with this pysat attribute. ',
                            '`user_modules` is not modifiable here.'))
            raise ValueError(estr)
        else:
            # General or user parameter, no additional processing
            self.data[key] = value

        # Store updated parameters to disk
        self.store()

    def _set_data_dirs(self, path=None, store=True):
        """
        Set the top level directories pysat uses to store and load data.

        Parameters
        ----------
        path : string or list-like of str
            Valid path(s) to directory
        store : bool
            Optionally store parameters to disk. Present to support a
            Deprecated method (default=True).

        """

        paths = np.asarray(path)
        if paths.shape == ():
            paths = [paths.tolist()]
        elif paths.shape[0] > 1:
            paths = paths.squeeze().tolist()
        elif paths.shape[0] == 1:
            paths = paths.tolist()

        # Account for a user prefix in the path, such as ~
        paths = [os.path.expanduser(path) for path in paths]

        # Account for the presence of $HOME or similar
        paths = [os.path.expandvars(path) for path in paths]

        # Make sure paths don't end with path separator for consistency
        paths = [path if path[-1] != os.path.sep else path[:-1]
                 for path in paths]

        # Ensure all paths are valid, create if not
        for path in paths:
            check_and_make_path(path)

        # Assign updated and validated paths
        self.data['data_dirs'] = paths

        # Optionally store information
        if store:
            self.store()

        return

    def clear_and_restart(self):
        """Clears all stored settings and sets pysat defaults

        pysat parameters without a default value are set to []
        """

        # Clear current data and assign a copy of default values
        self.data = copy.deepcopy(self.defaults)

        # Set pysat parameters without a default working value to []
        for key in self.non_defaults:
            self.data[key] = []

        # Trigger a file write
        self.store()

        return

    def restore_defaults(self):
        """Restore default pysat parameters

        Does not modify any stored custom user keys or pysat parameters
        without a default value.

        """

        # Set default values for each of the pysat provided values. Set
        # all but the last parameter directly. Set last using __setitem__
        # to trigger a file write.
        keys = list(self.defaults.keys())
        for key in keys:
            self.data[key] = self.defaults[key]

        # Trigger a file write
        self.store()

        return

    def store(self):
        """Store parameters using the filename specified in self.file_path.
        """

        # Store settings in file
        with Lock(self.file_path, 'w', self['file_timeout']) as fout:
            json.dump(self.data, fout)

            # Ensure write is fully complete even for network file systems
            fout.flush()
            os.fsync(fout.fileno())

        return



"""
tests the pysat meta object and code for instruments with ftp downloads.

Intended to be run locally, excluded from Travis CI
"""
from importlib import import_module
import os
import tempfile

import pysat
import pysat.instruments.pysat_testing
import pysat.tests.test_instruments as test_inst

include_list = ['icon_euv', 'icon_fuv', 'icon_ivm', 'icon_might', 'sw_dst']
# dict, keyed by pysat instrument, with a list of usernames and passwords
user_download_dict = {}


def init_func_external(self):
    """Iterate through and create all of the test Instruments needed.
       Only want to do this once.

    """

    # names of all the instrument modules
    instrument_names = pysat.instruments.__all__
    temp = []
    for name in instrument_names:
        if name in include_list:
            temp.append(name)
    instrument_names = temp
    self.instrument_names = temp
    self.instruments = []
    self.instrument_modules = []

    # create temporary directory
    dir_name = tempfile.mkdtemp()
    saved_path = pysat.data_dir
    pysat.utils.set_data_dir(dir_name, store=False)

    for name in instrument_names:
        try:
            print(' '.join(('FTP', name)))
            module = import_module(''.join(('.', name)),
                                   package='pysat.instruments')
        except ImportError:
            print("Couldn't import instrument module")
            pass
        else:
            # try and grab basic information about the module so we
            # can iterate over all of the options
            try:
                info = module._test_dates
            except AttributeError:
                info = {}
                info[''] = {'': pysat.datetime(2009, 1, 1)}
                module._test_dates = info
            for sat_id in info.keys():
                for tag in info[sat_id].keys():
                    try:
                        inst = pysat.Instrument(inst_module=module,
                                                tag=tag,
                                                sat_id=sat_id,
                                                temporary_file_list=True)
                        inst._test_dates = module._test_dates
                        self.instruments.append(inst)
                        self.instrument_modules.append(module)
                    except:
                        pass
    pysat.utils.set_data_dir(saved_path, store=False)


init_inst = None
init_mod = None
init_names = None

# this environment variable is set by the TRAVIS CI folks
if not (os.environ.get('TRAVIS') == 'true'):
    class TestFTPInstrumentQualifier(test_inst.TestInstrumentQualifier):

        def __init__(self):
            """Iterate through and create all of the test Instruments needed"""
            global init_inst
            global init_mod
            global init_names

            if init_inst is None:
                init_func_external(self)
                init_inst = self.instruments
                init_mod = self.instrument_modules
                init_names = self.instrument_names

            else:
                self.instruments = init_inst
                self.instrument_modules = init_mod
                self.instrument_names = init_names

        def setup(self):
            """Runs before every method to create a clean testing setup."""
            pass

        def teardown(self):
            """Runs after every method to clean up previous testing."""
            pass



"""
pysat.ssnl is a pysat module that provides
the interface to perform seasonal analysis on
data managed by pysat.  These analysis methods
are independent of instrument type.

Main Features
-------------
- Seasonal averaging routine for 1D and 2D data.
- Occurrence probability routines, daily or by orbit.
- Scatterplot of data_label(s) as functions of labelx,y
    over a season.

"""

from . import occur_prob
from . import avg
from . import plot
from ._core import computational_form



# -*- coding: utf-8 -*-
"""
Produces fake instrument data for testing.
"""

import datetime as dt
import functools
import numpy as np

import xarray as xr

import pysat
from pysat.instruments.methods import testing as mm_test

logger = pysat.logger

platform = 'pysat'
name = 'testing2d_xarray'

tags = {'': 'Regular testing data set'}
sat_ids = {'': ['']}
pandas_format = False
tags = {'': 'Regular testing data set'}
inst_ids = {'': ['']}
_test_dates = {'': {'': dt.datetime(2009, 1, 1)}}

epoch_name = u'time'


# Init method
init = mm_test.init


# Clean method
clean = mm_test.clean


# Optional method, preprocess
preprocess = mm_test.preprocess


def load(fnames, tag=None, inst_id=None, malformed_index=False,
         num_samples=None, test_load_kwarg=None):
    """ Loads the test files

    Parameters
    ----------
    fnames : list
        List of filenames
    tag : str or NoneType
        Instrument tag (accepts '')
    inst_id : str or NoneType
        Instrument satellite ID (accepts '')
    malformed_index : bool False
        If True, the time index will be non-unique and non-monotonic.
    num_samples : int
        Number of samples
    test_load_kwarg : any or NoneType
        Testing keyword (default=None)

    Returns
    -------
    data : xr.Dataset
        Testing data
    meta : pysat.Meta
        Testing metadata

    """

    # Support keyword testing
    logger.info(''.join(('test_load_kwarg = ', str(test_load_kwarg))))

    # create an artifical satellite data set
    iperiod = mm_test.define_period()
    drange = mm_test.define_range()

    if num_samples is None:
        # Default to 1 day at a frequency of 100S
        num_samples = 864
    # Using 100s frequency for compatibility with seasonal analysis unit tests
    uts, index, dates = mm_test.generate_times(fnames, num_samples,
                                               freq='100S')

    if malformed_index:
        index = index.tolist()
        # nonmonotonic
        index[0:3], index[3:6] = index[3:6], index[0:3]
        # non unique
        index[6:9] = [index[6]] * 3
    data = xr.Dataset({'uts': ((epoch_name), index)},
                      coords={epoch_name: index})

    # need to create simple orbits here. Have start of first orbit
    # at 2009,1, 0 UT. 14.84 orbits per day
    # figure out how far in time from the root start
    # use that info to create a signal that is continuous from that start
    # going to presume there are 5820 seconds per orbit (97 minute period)
    time_delta = dates[0] - dt.datetime(2009, 1, 1)

    # mlt runs 0-24 each orbit.
    mlt = mm_test.generate_fake_data(time_delta.total_seconds(), uts,
                                     period=iperiod['lt'],
                                     data_range=drange['lt'])
    data['mlt'] = ((epoch_name), mlt)

    # do slt, 20 second offset from mlt
    slt = mm_test.generate_fake_data(time_delta.total_seconds() + 20, uts,
                                     period=iperiod['lt'],
                                     data_range=drange['lt'])
    data['slt'] = ((epoch_name), slt)

    # create a fake satellite longitude, resets every 6240 seconds
    # sat moves at 360/5820 deg/s, Earth rotates at 360/86400, takes extra time
    # to go around full longitude
    longitude = mm_test.generate_fake_data(time_delta.total_seconds(), uts,
                                           period=iperiod['lon'],
                                           data_range=drange['lon'])
    data['longitude'] = ((epoch_name), longitude)

    # create fake satellite latitude for testing polar orbits
    angle = mm_test.generate_fake_data(time_delta.total_seconds(), uts,
                                       period=iperiod['angle'],
                                       data_range=drange['angle'])
    latitude = 90.0 * np.cos(angle)
    data['latitude'] = ((epoch_name), latitude)

    # create constant altitude at 400 km for a satellite that has yet
    # to experience orbital decay
    alt0 = 400.0
    altitude = alt0 * np.ones(data['latitude'].shape)
    data['altitude'] = ((epoch_name), altitude)

    # create some fake data to support testing of averaging routines
    mlt_int = data['mlt'].astype(int).data
    long_int = (data['longitude'] / 15.).astype(int).data
    data['dummy1'] = ((epoch_name), mlt_int)
    data['dummy2'] = ((epoch_name), long_int)
    data['dummy3'] = ((epoch_name), mlt_int + long_int * 1000.)
    data['dummy4'] = ((epoch_name), uts)

    # Add dummy coords
    data.coords['x'] = (('x'), np.arange(17))
    data.coords['y'] = (('y'), np.arange(17))
    data.coords['z'] = (('z'), np.arange(15))

    # create altitude 'profile' at each location to simulate remote data
    num = len(data['uts'])
    data['profiles'] = (
        (epoch_name, 'profile_height'),
        data['dummy3'].values[:, np.newaxis] * np.ones((num, 15)))
    data.coords['profile_height'] = ('profile_height', np.arange(15))

    # profiles that could have different altitude values
    data['variable_profiles'] = (
        (epoch_name, 'z'), data['dummy3'].values[:, np.newaxis]
        * np.ones((num, 15)))
    data.coords['variable_profile_height'] = (
        (epoch_name, 'z'), np.arange(15)[np.newaxis, :] * np.ones((num, 15)))

    # Create fake image type data, projected to lat / lon at some location
    # from satellite
    data['images'] = ((epoch_name, 'x', 'y'),
                      data['dummy3'].values[
                          :, np.newaxis, np.newaxis] * np.ones((num, 17, 17)))
    data.coords['image_lat'] = \
        ((epoch_name, 'x', 'y'),
         np.arange(17)[np.newaxis,
                       np.newaxis,
                       :] * np.ones((num, 17, 17)))
    data.coords['image_lon'] = ((epoch_name, 'x', 'y'),
                                np.arange(17)[np.newaxis, np.newaxis,
                                              :] * np.ones((num, 17, 17)))

    # create very limited metadata
    meta = pysat.Meta()
    meta[epoch_name] = {'long_name': 'Datetime Index'}
    meta['uts'] = {'units': 's', 'long_name': 'Universal Time'}
    meta['mlt'] = {'units': 'hours', 'long_name': 'Magnetic Local Time'}
    meta['slt'] = {'units': 'hours', 'long_name': 'Solar Local Time'}
    meta['longitude'] = {'units': 'degrees', 'long_name': 'Longitude'}
    meta['latitude'] = {'units': 'degrees', 'long_name': 'Latitude'}
    meta['altitude'] = {'units': 'km', 'long_name': 'Altitude'}
    variable_profile_meta = pysat.Meta()
    variable_profile_meta['variable_profiles'] = {'long_name': 'series'}
    meta['variable_profiles'] = {'meta': variable_profile_meta,
                                 'long_name': 'series'}
    profile_meta = pysat.Meta()
    profile_meta['density'] = {'long_name': 'profiles'}
    profile_meta['dummy_str'] = {'long_name': 'profiles'}
    profile_meta['dummy_ustr'] = {'long_name': 'profiles'}
    meta['profiles'] = {'meta': profile_meta, 'long_name': 'profiles'}
    image_meta = pysat.Meta()
    image_meta['density'] = {'long_name': 'profiles'}
    image_meta['fraction'] = {'long_name': 'profiles'}
    meta['images'] = {'meta': image_meta, 'long_name': 'profiles'}
    for var in data.keys():
        if var.find('dummy') >= 0:
            meta[var] = {'units': 'none', 'long_name': var,
                         'notes': 'Dummy variable'}
    meta['x'] = {'long_name': 'x-value of image pixel',
                 'notes': 'Dummy Variable'}
    meta['y'] = {'long_name': 'y-value of image pixel',
                 'notes': 'Dummy Variable'}
    meta['z'] = {'long_name': 'z-value of profile height',
                 'notes': 'Dummy Variable'}
    meta['image_lat'] = {'long_name': 'Latitude of image pixel',
                         'notes': 'Dummy Variable'}
    meta['image_lon'] = {'long_name': 'Longitude of image pixel',
                         'notes': 'Dummy Variable'}
    meta['profile_height'] = {'long_name': 'profile height'}
    meta['variable_profile_height'] = {'long_name': 'Variable Profile Height'}

    return data, meta


list_files = functools.partial(mm_test.list_files, test_dates=_test_dates)
list_remote_files = functools.partial(mm_test.list_remote_files,
                                      test_dates=_test_dates)
download = functools.partial(mm_test.download)



# -*- coding: utf-8 -*-
"""Supports the SEE instrument on TIMED.

Downloads data from the NASA Coordinated Data
Analysis Web (CDAWeb).

Supports two options for loading that may be
specified at instantiation.

Properties
----------
platform
    'timed'
name
    'see'
tag
    None
sat_id
    None supported
flatten_twod
    If True, then two dimensional data is flattened across
    columns. Name mangling is used to group data, first column
    is 'name', last column is 'name_end'. In between numbers are
    appended 'name_1', 'name_2', etc. All data for a given 2D array
    may be accessed via, data.loc[:, 'item':'item_end']
    If False, then 2D data is stored as a series of DataFrames,
    indexed by Epoch. data.loc[0, 'item']
    (default=True)


Note
----
- no tag required


Warnings
--------
- Currently no cleaning routine.

"""

from __future__ import print_function
from __future__ import absolute_import
import functools

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

# include basic instrument info
platform = 'timed'
name = 'see'
tags = {'': ''}
sat_ids = {'': ['']}
_test_dates = {'': {'': pysat.datetime(2009, 1, 1)}}


# support list files routine
# use the default CDAWeb method
fname = 'timed_l3a_see_{year:04d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'': fname}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags,
                               fake_daily_files_from_monthly=True)

# support download routine
# use the default CDAWeb method
basic_tag = {'dir': '/pub/data/timed/see/data/level3a_cdf',
             'remote_fname': '{year:4d}/{month:02d}/'+fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)
# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)

# support load routine
# use the default CDAWeb method
load = functools.partial(cdw.load, fake_daily_files_from_monthly=True)


def clean(inst):
    """Routine to return TIMED SEE data cleaned to the specified level

    Parameters
    ----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Note
    ----
    No cleaning currently available.

    """

    return None



# -*- coding: utf-8 -*-
"""This is a template for a pysat.Instrument support file that
utilizes CDAWeb methods. Copy and modify this file as needed when adding a
new Instrument to pysat.

This is a good area to introduce the instrument, provide background
on the mission, operations, instrumenation, and measurements.

Also a good place to provide contact information. This text will
be included in the pysat API documentation.

Properties
----------
platform
    *List platform string here*
name
    *List name string here*
sat_id
    *List supported sat_ids here*
tag
    *List supported tag strings here*

Note
----
- Optional section, remove if no notes


Warnings
--------
- Optional section, remove if no warnings
- Two blank lines needed afterward for proper formatting


Examples
--------
::

    Example code can go here


Authors
-------
Author name and institution

"""

from __future__ import print_function
from __future__ import absolute_import

import functools
import numpy as np
import sys
import pandas as pds
import pysat

# CDAWeb methods prewritten for pysat
from pysat.instruments.methods import nasa_cdaweb as cdw

# the platform and name strings associated with this instrument
# need to be defined at the top level
# these attributes will be copied over to the Instrument object by pysat
# the strings used here should also be used to name this file
# platform_name.py
platform = ''
name = ''

# dictionary of data 'tags' and corresponding description
tags = {'': 'description 1',  # this is the default
        'tag_string': 'description 2'}

# Let pysat know if there are multiple satellite platforms supported
# by these routines
# define a dictionary keyed by satellite ID, each with a list of
# corresponding tags
# sat_ids = {'a':['L1', 'L0'], 'b':['L1', 'L2'], 'c':['L1', 'L3']}
sat_ids = {'': ['']}

# Define good days to download data for when pysat undergoes testing.
# format is outer dictionary has sat_id as the key
# each sat_id has a dictionary of test dates keyed by tag string
# _test_dates = {'a':{'L0':pysat.datetime(2019,1,1),
#                     'L1':pysat.datetime(2019,1,2)},
#                'b':{'L1':pysat.datetime(2019,3,1),
#                     'L2':pysat.datetime(2019,11,23),}}
_test_dates = {'': {'': pysat.datetime(2019, 1, 1)}}

# Additional information needs to be defined
# to support the CDAWeb list files routine
# We need to define a filename format string for every
# supported combination of sat_id and tag string
# fname1 = 'cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf'
# fname2 = 'cnofs_vefi_acfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf'
# supported_tags = {'sat1':{'tag1':fname1},
#                   'sat2':{'tag2':fname2}}
# you can use format keywords year, month, day, hour, min, sec,
# version and revision
# see code docstring for latest
fname = 'cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf'
supported_tags = {'': {'': fname}}
# use the CDAWeb methods list files routine
# the command below presets some of the methods inputs, leaving
# those provided by pysat available when invoked
list_files = functools.partial(cdw.list_files,
                               supported_tags=supported_tags)

#
# support load routine
#
# use the default CDAWeb method
# no other information needs to be supplied here
# pysatCDF is used to load data
load = cdw.load

#
# support download routine
#
# to use the default CDAWeb method
# we need to provide additional information
# directory location on CDAWeb ftp server
# formatting template for filenames on CDAWeb
# formatting template for files saved to the local disk
# a dictionary needs to be created for each sat_id and tag
# combination along with the file format template
# outer dict keyed by sat_id, inner dict keyed by tag
basic_tag = {'dir': '/pub/data/cnofs/vefi/bfield_1sec',
             'remote_fname': '{year:4d}/' + fname,
             'local_fname': fname}
supported_tags = {'': {'': basic_tag}}
download = functools.partial(cdw.download, supported_tags)

# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


# code should be defined below as needed
def default(self):
    """Default customization function.

    This routine is automatically applied to the Instrument object
    on every load by the pysat nanokernel (first in queue).

    Parameters
    ----------
    self : pysat.Instrument
        This object

    """

    return


# code should be defined below as needed
def clean(inst):
    """Routine to return PLATFORM/NAME data cleaned to the specified level

    Cleaning level is specified in inst.clean_level and pysat
    will accept user input for several strings. The clean_level is
    specified at instantiation of the Instrument object.

    'clean' All parameters should be good, suitable for statistical and
            case studies
    'dusty' All paramers should generally be good though same may
            not be great
    'dirty' There are data areas that have issues, data should be used
            with caution
    'none'  No cleaning applied, routine not called in this case.


    Parameters
    -----------
    inst : pysat.Instrument
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    """

    return



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
""" Utilities to perform common evaluations
"""

import numpy as np


def assert_list_contains(small_list, big_list):
    """ Assert all elements of one list exist within the other list

    Parameters
    ----------
    small_list : list
        List whose values must all be present within big_list
    big_list : list
        List that must contain all the values in small_list

    Raises
    ------
    AssertionError
        If a small_list value is missing from big_list

    """

    for value in small_list:
        assert value in big_list, "{:} not in {:}".format(value, big_list)

    return


def assert_lists_equal(list1, list2):
    """Assert that the lists contain the same elements

    Parameters
    ----------
    list1 : list
        Input list one
    list2 : list
        Input list two

    Raises
    ------
    AssertionError
        If a list1 value is missing from list2 or list lengths are unequal

    Note
    ----
    This test does not require that the lists have the same elements in the
    same order, and so is also a good test for keys.

    """

    assert len(list1) == len(list2)
    assert_list_contains(list1, list2)

    return


def nan_equal(value1, value2):
    """ Determine if values are equal or are both NaN

    Parameters
    ----------
    value1 : scalar-like
        Value of any type that can be compared without iterating
    value2 : scalar-like
        Another value of any type that can be compared without iterating

    Returns
    -------
    is_equal : bool
        True if both values are equal or NaN, False if they are not

    """

    is_equal = (value1 == value2)

    if not is_equal:
        try:
            if np.isnan(value1) and np.isnan(value2):
                is_equal = True
        except TypeError:
            # One or both of value1 and value2 cannot be evaluated by np.isnan
            # and so have been correctly identified as unequal
            pass

    return is_equal
